from django.urls import reverse
from arkindex_common.ml_tool import MLToolType
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex.documents.models import Corpus, DataSource, MLClass
from arkindex.dataimport.models import WorkerVersion


class TestBulkClassification(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 2r')
        cls.src = DataSource.objects.get(slug='test', type=MLToolType.Classifier)
        cls.private_corpus = Corpus.objects.create(name='private', public=False)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.src.internal = True
        cls.src.save()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()
        cls.src.internal = False
        cls.src.save()

    def create_classifications_data(self, classifications, parent=None):
        return {
            "parent": parent or str(self.page.id),
            "worker_version": str(self.worker_version.id),
            "classifications": classifications
        }

    def test_bulk_classification_requires_login(self):
        response = self.client.post(reverse('api:classification-bulk'), format='json')
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_bulk_classification_wrong_acl(self):
        self.client.force_login(self.user)
        private_page = self.private_corpus.elements.create(
            type=self.private_corpus.types.create(slug='page'),
        )
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data(
                    [{"class_name": 'dog', "confidence": 0.99}],
                    parent=str(private_page.id)
                )
            )

        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(
            response.json(),
            {
                'parent': ['Invalid pk "{}" - object does not exist.'.format(private_page.id)]
            }
        )

    def test_bulk_classification_requires_source_xor_worker_version(self):
        """
        A classifier data source XOR a worker_version is required to push classifications on an element
        """
        wrong_payloads = (
            {'classifier': self.src.slug, 'worker_version': self.worker_version.id},
            {'classifier': None, 'worker_version': None},
            {'classifier': ''},
            {}
        )
        self.client.force_login(self.user)
        for payload in wrong_payloads:
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    'parent': str(self.page.id),
                    'classifications': [{'class_name': 'cat', 'confidence': 0.42}],
                    **payload
                }
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            if payload.get('classifier') and payload.get('worker_version'):
                self.assertDictEqual(response.json(), {
                    'classifier': ['This field XOR worker_version field must be set to create classifications'],
                    'worker_version': ['This field XOR classifier field must be set to create classifications']
                })

    def test_bulk_classification_source(self):
        """
        Bulk classifications are created using an existing classifier source
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data={
                    "parent": str(self.page.id),
                    "classifier": self.src.slug,
                    "classifications": [{
                        "class_name": 'dog',
                        "confidence": 0.99,
                        "high_confidence": True
                    }, {
                        "class_name": 'cat',
                        "confidence": 0.42,
                    }]
                }
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(
            list(self.page.classifications.values_list(
                'ml_class__name',
                'confidence',
                'high_confidence',
                'source',
                'worker_version'
            )),
            [
                ('dog', 0.99, True, self.src.id, None),
                ('cat', 0.42, False, self.src.id, None)
            ],
        )

    def test_bulk_classification_worker_version(self):
        """
        Classifications are created and linked to a worker version
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data([
                    {
                        "class_name": 'dog',
                        "confidence": 0.99,
                        "high_confidence": True
                    },
                    {
                        "class_name": 'cat',
                        "confidence": 0.42,
                    }
                ])
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(
            list(self.page.classifications.values_list(
                'ml_class__name',
                'confidence',
                'high_confidence',
                'source',
                'worker_version'
            )),
            [
                ('dog', 0.99, True, None, self.worker_version.id),
                ('cat', 0.42, False, None, self.worker_version.id)
            ],
        )

    def test_bulk_classification_create_ml_class(self):
        """
        Adding classifications with non existing classes should automatically create them
        """
        self.assertEqual(MLClass.objects.filter(name="dog", corpus=self.corpus).count(), 0)
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data(
                    [{"class_name": 'dog', "confidence": 0.99}],
                )
            )
        self.assertEqual(MLClass.objects.filter(name="dog", corpus=self.corpus).count(), 1)

    def test_bulk_classification_existing_ml_class(self):
        MLClass.objects.create(name="cat", corpus=self.corpus)
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data(
                    [{"class_name": 'cat', "confidence": 0.42}],
                )
            )
        self.assertEqual(MLClass.objects.filter(name="cat", corpus=self.corpus).count(), 1)

    def test_bulk_classification_delete(self):
        """
        Test the bulk classification API deletes previous classifications with a similar worker version
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data([
                    {
                        "class_name": 'dog',
                        "confidence": 0.99,
                        "high_confidence": True
                    },
                    {
                        "class_name": 'cat',
                        "confidence": 0.42,
                    }
                ])
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data([
                    {
                        "class_name": 'doggo',
                        "confidence": 0.5,
                    },
                    {
                        "class_name": 'catte',
                        "confidence": 0.85,
                        "high_confidence": True
                    }
                ])
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertCountEqual(
            list(self.page.classifications.values_list('ml_class__name', 'confidence', 'high_confidence')),
            [
                ('doggo', 0.5, False),
                ('catte', 0.85, True),
            ],
        )

    def test_bulk_classification_no_duplicates(self):
        """
        Test the bulk classification API prevents creating classifications with duplicate ML classes
        """
        self.client.force_login(self.user)
        with self.assertNumQueries(4):
            response = self.client.post(
                reverse('api:classification-bulk'),
                format='json',
                data=self.create_classifications_data([
                    {"class_name": 'dog', "confidence": 0.99},
                    {"class_name": 'dog', "confidence": 0.99},
                ])
            )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'classifications': ['Duplicated ML classes are not allowed from the same source or worker version.']
        })
