from django.urls import reverse
from django.test import override_settings
from django.db.models import Count
from unittest.mock import patch
from mock import AsyncMock
from rest_framework import status
from arkindex.project.tests import FixtureAPITestCase
from arkindex_common.enums import TranscriptionType
from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import Element, Corpus, DataSource
from arkindex_common.ml_tool import MLToolType


class TestBulkElementTranscriptions(FixtureAPITestCase):
    """
    Tests for text element creation view
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page = cls.corpus.elements.get(name='Volume 1, page 1r')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.src = DataSource.objects.get(slug='test', type=MLToolType.Recognizer)
        cls.line = cls.corpus.elements.filter(type__slug='text_line').first()
        cls.private_corpus = Corpus.objects.create(name='Private')
        cls.private_page = cls.private_corpus.elements.create(type=cls.page.type)
        cls.worker_version = WorkerVersion.objects.get(worker__slug='reco')

    def setUp(self):
        self.manual_source = DataSource.objects.create(type=MLToolType.Recognizer, slug='manual', internal=False)
        self.manual_transcription = self.line.transcriptions.create(
            text='A manual transcription',
            source=self.manual_source,
            type=TranscriptionType.Line
        )
        self.private_transcription = self.private_page.transcriptions.create(
            text='PEPE',
            type=TranscriptionType.Line,
            source=self.manual_source
        )

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_bulk_transcriptions_with_source(self, get_layer_mock):
        """
        Bulk creates a list of element with an attached transcription generated by a source
        """
        get_layer_mock.return_value.send = AsyncMock()
        self.src.internal = True
        self.src.save()

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'Hello world !', 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], 'I <3 JavaScript', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'transcription_type': 'line',
            'source': self.src.slug,
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'score': score
            } for poly, text, score in transcriptions]
        }

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(17):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id=self.line.id)
        self.assertEqual(created_elts.count(), 2)
        self.assertSetEqual(
            set(created_elts.values_list('zone__image_id', flat=True)),
            {self.page.zone.image_id}
        )
        self.assertListEqual(
            [
                (elt.paths.get().ordering, elt.name, elt.zone.polygon.coords)
                for elt in created_elts.order_by('name')
            ],
            [
                (1, '2', ((13, 37), (13, 137), (133, 137), (133, 37), (13, 37))),
                (2, '3', ((24, 42), (24, 142), (64, 142), (64, 42), (24, 42)))
            ]
        )
        self.assertCountEqual(
            created_elts.values_list('transcriptions__type', 'transcriptions__text', 'transcriptions__zone', 'transcriptions__source', 'transcriptions__worker_version'),
            [
                (TranscriptionType.Line, ('Hello world !'), None, self.src.id, None),
                (TranscriptionType.Line, ('I <3 JavaScript'), None, self.src.id, None)
            ]
        )
        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_bulk_transcriptions_with_worker_version(self, get_layer_mock):
        """
        Bulk creates a list of element with an attached transcription generated by a worker_version
        """
        get_layer_mock.return_value.send = AsyncMock()
        self.src.internal = True
        self.src.save()

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'Hello world !', 0.1337),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], 'I <3 JavaScript', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'transcription_type': 'line',
            'worker_version': str(self.worker_version.id),
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'score': score
            } for poly, text, score in transcriptions]
        }
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data=data
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id=self.line.id)
        self.assertEqual(created_elts.count(), 2)
        self.assertTrue(all(map(lambda elt: elt.zone.image == self.page.zone.image, created_elts)))
        self.assertListEqual(
            [
                (elt.paths.first().ordering, elt.name, elt.zone.polygon.coords)
                for elt in created_elts
            ],
            [
                (1, '2', ((13, 37), (13, 137), (133, 137), (133, 37), (13, 37))),
                (2, '3', ((24, 42), (24, 142), (64, 142), (64, 42), (24, 42)))
            ]
        )
        self.assertCountEqual(
            created_elts.values_list('transcriptions__type', 'transcriptions__text', 'transcriptions__zone', 'transcriptions__source', 'transcriptions__worker_version'),
            [
                (TranscriptionType.Line, ('Hello world !'), None, None, self.worker_version.id),
                (TranscriptionType.Line, ('I <3 JavaScript'), None, None, self.worker_version.id)
            ]
        )
        get_layer_mock().send.assert_called_once_with('reindex', {
            'type': 'reindex.start',
            'corpus': None,
            'element': str(self.page.id),
            'entity': None,
            'transcriptions': True,
            'elements': True,
            'entities': False,
            'drop': False,
        })

    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_bulk_transcriptions_load(self):
        """
        Sending 100 transcriptions should not increase the number of DB queries
        """
        self.src.internal = True
        self.src.save()

        transcriptions = [(
            [[i, i], [i, i + 20], [i + 20, i + 20], [i + 20, i], [i, i]],
            str(i / 2),
            i / 200
        ) for i in range(1, 200, 2)]
        data = {
            'element_type': 'text_line',
            'transcription_type': 'line',
            'source': self.src.slug,
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'score': score
            } for poly, text, score in transcriptions]
        }

        created_elts = Element.objects.get_descending(self.page.id) \
            .filter(paths__path__last=self.page.id, type__slug='text_line')
        self.assertEqual(created_elts.count(), 1)

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(17):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(created_elts.count(), 101)
        # Each element has a transcription
        self.assertEqual(created_elts.annotate(ts_count=Count('transcriptions')).filter(ts_count=1).count(), 101)

    @patch('arkindex.project.triggers.get_channel_layer')
    def test_bulk_transcriptions_similar_zone(self, get_layer_mock):
        """
        Pushing a transcription matching an element type and zone reuses this element
        """
        get_layer_mock.return_value.send = AsyncMock()
        self.src.internal = True
        self.src.save()

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'Hello world !', 0.1337),
            (list(self.line.zone.polygon), 'I <3 JavaScript', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'transcription_type': 'line',
            'source': self.src.slug,
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'score': score
            } for poly, text, score in transcriptions]
        }

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(17):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id)
        # The existing text line has been reused
        self.assertEqual(created_elts.count(), 2)
        self.assertCountEqual(self.line.transcriptions.values_list('type', 'text', 'zone'), [
            (TranscriptionType.Line, 'A manual transcription', None),
            (TranscriptionType.Line, 'I <3 JavaScript', None)
        ])

    def test_bulk_transcriptions_requires_verified(self):
        """
        The user has to be verified tu push transcriptions in bulk
        """
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={}
        )
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)
        self.assertDictEqual(response.json(), {'detail': 'You do not have permission to perform this action.'})

    def test_bulk_transcriptions_private(self):
        """
        Accessing a non writeable element triggers a 404_NOT_FOUND
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.private_page.id}),
            format='json',
            data={}
        )
        self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    def test_bulk_transcriptions_no_source_no_worker_version(self):
        """
        Transcriptions must be created with a source or a worker_version
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={
                'element_type': 'text_line',
                'transcription_type': 'line',
                'transcriptions': [{
                    'polygon': [[13, 37], [133, 37], [133, 137], [13, 137]],
                    'text': 'Can I write womething here ?',
                    'score': .666
                }]
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'source': ['This field XOR worker_version field must be set to create transcriptions'],
            'worker_version': ['This field XOR source field must be set to create transcriptions']
        })

    def test_bulk_transcriptions_with_source_and_worker_version_returns_error(self):
        """
        Transcriptions must be created with a source or a worker_version not both
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={
                'element_type': 'text_line',
                'transcription_type': 'line',
                'source': self.src.slug,
                'worker_version': str(self.worker_version.id),
                'transcriptions': [{
                    'polygon': [[13, 37], [133, 37], [133, 137], [13, 137]],
                    'text': 'Can I write womething here ?',
                    'score': .666
                }]
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'source': ['You can only refer to a DataSource XOR a WorkerVersion on transcriptions'],
            'worker_version': ['You can only refer to a DataSource XOR a WorkerVersion on transcriptions']
        })

    def test_bulk_transcriptions_non_manual(self):
        """
        Transcriptions source cannot be set to manual
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={
                'element_type': 'text_line',
                'transcription_type': 'line',
                'source': 'manual',
                'transcriptions': [{
                    'polygon': [[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]],
                    'text': 'Can I write womething here ?',
                    'score': .666
                }]
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'source': ['Transcriptions source slug cannot be set to manual.']
        })

    def test_bulk_transcriptions_wrong_fields_validation(self):
        """
        Briefly assert fields values are validated
        """
        self.client.force_login(self.user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
            format='json',
            data={
                'element_type': 'paragraph',
                'transcription_type': 'boulga',
                'source': 'gloubiboulga',
                'transcriptions': [{
                    'polygon': [[13, 37]],
                    'text': 'There is a snake in my computer'
                }]
            }
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'element_type': ['Object with slug=paragraph does not exist.'],
            'source': ["Source with slug 'gloubiboulga' not found"],
            'transcription_type': ['Value is not of type TranscriptionType'],
            'transcriptions': [{
                'polygon': ['Ensure this field has at least 3 elements.'],
                'score': ['This field is required.']
            }]
        })

    @override_settings(ARKINDEX_FEATURES={'search': False})
    def test_bulk_transcriptions_return_ordered_elements(self):
        """
        return_elements field allows to serialize an ordered list of sub-elements holding the transcriptions
        Returned elements may be duplicated in case the same zone is used for multiple transcriptions
        """
        self.src.internal = True
        self.src.save()

        transcriptions = [
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'stuck', 0.13),
            ([[13, 37], [133, 37], [133, 137], [13, 137], [13, 37]], 'stick', 0.33),
            ([[24, 42], [64, 42], [64, 142], [24, 142], [24, 42]], 'stock', 0.37),
            # The existing line has the smallest ordering
            (self.line.zone.polygon.coords, 'stack', 0.42),
        ]
        data = {
            'element_type': 'text_line',
            'transcription_type': 'line',
            'source': self.src.slug,
            'transcriptions': [{
                'polygon': poly,
                'text': text,
                'score': score
            } for poly, text, score in transcriptions],
            'return_elements': True
        }

        self.client.force_login(self.internal_user)
        with self.assertNumQueries(17):
            response = self.client.post(
                reverse('api:element-transcriptions-bulk', kwargs={'pk': self.page.id}),
                format='json',
                data=data
            )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        created_elts = Element.objects.get_descending(self.page.id).exclude(id=self.line.id)
        self.assertEqual(created_elts.count(), 2)

        first_elt = created_elts.get(transcriptions__text='stuck')
        second_elt = created_elts.get(transcriptions__text='stock')
        self.assertListEqual(response.json(), [{
            'id': str(first_elt.id),
            'created': True
        }, {
            'id': str(first_elt.id),
            'created': False
        }, {
            'id': str(second_elt.id),
            'created': True
        }, {
            'id': str(self.line.id),
            'created': False
        }])

    def test_bulk_transcriptions_requires_zone(self):
        self.client.force_login(self.internal_user)
        response = self.client.post(
            reverse('api:element-transcriptions-bulk', kwargs={'pk': self.vol.id}),
            format='json',
            data={}
        )
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {'element': ['Cannot create transcriptions on an element without a zone.']})
