from django.urls import reverse
from rest_framework import status
from arkindex_common.ml_tool import MLToolType
from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import \
    Element, DataSource, Corpus
from arkindex.images.models import ImageServer
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase


class TestCreateElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug='volume')
        cls.page_type = cls.corpus.types.get(slug='page')
        cls.act_type = cls.corpus.types.get(slug='act')
        cls.vol = cls.corpus.elements.get(name='Volume 1')
        cls.element = Element.objects.get(name='Volume 1, page 2r')
        cls.image = ImageServer.objects.local.images.create(
            path="kingdom/far/away",
            status=S3FileStatus.Checked,
            width=42,
            height=42,
        )
        cls.worker_version = WorkerVersion.objects.get(worker__slug='dla')

    def make_create_request(self, name='default', corpus=None, elt_type='volume', **options):
        request = {
            'path': reverse('api:elements-create'),
            'data': {
                'type': elt_type,
                'name': name,
                'corpus': corpus or str(self.corpus.id),
            },
            'format': 'json',
        }
        # Optional parameters
        request['data'].update(options)
        return request

    def test_create_element_not_verified(self):
        self.assertEqual(
            self.client.post(**self.make_create_request('no')).status_code,
            status.HTTP_403_FORBIDDEN
        )

    def test_create_element_volume(self):
        # Create a Volume
        self.client.force_login(self.user)
        request = self.make_create_request('my new volume')
        with self.assertNumQueries(7):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        data = response.json()
        self.assertIn('id', data)
        volume = Element.objects.get(id=data['id'])
        # Assert the fully serialized response matches the created volume
        self.assertDictEqual(
            data,
            {
                'id': str(volume.id),
                'name': volume.name,
                'type': volume.type.slug,
                'thumbnail_put_url': None,
                'thumbnail_url': volume.thumbnail.s3_url,
                'source': None,
                'worker_version': None,
                'classifications': [],
                'metadata': [],
                'corpus': {
                    'id': str(volume.corpus.id),
                    'name': volume.corpus.name,
                    'public': volume.corpus.public
                },
                'zone': None
            }
        )
        self.assertEqual(volume.name, 'my new volume')
        self.assertEqual(volume.type, self.volume_type)

    def test_create_volume_page(self):
        # Add a page with an Image to an existing volume
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams',
            image=str(self.image.id),
        )
        with self.assertNumQueries(16):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        # Assert the fully serialized response matches the created page (exclude image zone)
        del data['zone']
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'name': page.name,
                'type': page.type.slug,
                'thumbnail_put_url': None,
                'thumbnail_url': None,
                'source': None,
                'worker_version': None,
                'classifications': [],
                'metadata': [],
                'corpus': {
                    'id': str(page.corpus.id),
                    'name': page.corpus.name,
                    'public': page.corpus.public
                }
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams')
        self.assertEqual(page.type, self.page_type)

    def test_create_volume_act(self):
        # Create an volume child of type Act
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            name='Castle story',
            elt_type='act'
        )
        with self.assertNumQueries(13):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)

    def test_create_element_source(self):
        # Create an element with a source
        self.client.force_login(self.user)
        source = DataSource.objects.create(
            type=MLToolType.DLAnalyser,
            slug='fairy_tale_detector',
            internal=False,
        )
        request = self.make_create_request(
            name='Castle story',
            elt_type='act',
            source='fairy_tale_detector',
        )
        with self.assertNumQueries(8):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.source, source)
        self.assertEqual(act.worker_version, None)

    def test_create_element_worker_version(self):
        # Create an element with a worker version
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='Castle story',
            elt_type='act',
            worker_version=str(self.worker_version.id),
        )
        with self.assertNumQueries(8):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()['id'])
        self.assertEqual(act.name, 'Castle story')
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.source, None)
        self.assertEqual(act.worker_version, self.worker_version)

    def test_create_element_source_and_worker_version_returns_error(self):
        # Create an element with a source and a worker version (not allowed)
        self.client.force_login(self.user)
        DataSource.objects.create(
            type=MLToolType.DLAnalyser,
            slug='fairy_tale_detector',
            internal=False,
        )
        request = self.make_create_request(
            name='Castle story',
            elt_type='act',
            source='fairy_tale_detector',
            worker_version=str(self.worker_version.id),
        )
        with self.assertNumQueries(6):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'source': ['You can only refer to a DataSource XOR a WorkerVersion on an '
                       'element'],
            'worker_version': ['You can only refer to a DataSource XOR a WorkerVersion on '
                               'an element']
        })

    def test_create_element_polygon(self):
        # Create an element with a polygon to an existing volume
        polygon = [[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]]
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            name='The castle of my dreams again',
            image=str(self.image.id),
            polygon=polygon
        )
        with self.assertNumQueries(16):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data['id'])
        # Assert the fully serialized response matches the created element (exclude zone image field)
        del data['zone']['image']
        self.assertDictEqual(
            data,
            {
                'id': str(page.id),
                'name': page.name,
                'type': page.type.slug,
                'thumbnail_put_url': None,
                'thumbnail_url': None,
                'source': None,
                'worker_version': None,
                'classifications': [],
                'metadata': [],
                'corpus': {
                    'id': str(page.corpus.id),
                    'name': page.corpus.name,
                    'public': page.corpus.public
                },
                'zone': {
                    'id': str(page.zone.id),
                    'polygon': polygon,
                    'url': str(page.zone.url)
                }
            }
        )
        self.assertEqual(page.name, 'The castle of my dreams again')
        self.assertEqual(page.type, self.page_type)

    def test_create_parent_different_corpus(self):
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        new_corpus.corpus_right.create(user=self.user, can_write=True)
        new_corpus.types.create(slug='act')
        request = self.make_create_request(
            parent=str(self.vol.id),
            corpus=str(new_corpus.id),
            elt_type='act'
        )
        with self.assertNumQueries(5):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'corpus': ['Parent and child must be in the same corpus']
        })

    def test_create_filter_type(self):
        """
        Ensure types are filtered by corpus
        """
        self.assertTrue(self.corpus.types.filter(slug='volume').exists())
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name='new')
        new_corpus.corpus_right.create(user=self.user, can_write=True)
        new_vol_type = new_corpus.types.create(slug='volume', display_name='Volume', folder=True)
        request = self.make_create_request(
            corpus=str(new_corpus.id),
            elt_type='volume',
            name='something',
        )
        with self.assertNumQueries(7):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_vol = new_corpus.elements.get()
        self.assertEqual(new_vol.type, new_vol_type)
        self.assertEqual(new_vol.type.corpus, new_corpus)

    def test_create_wrong_type(self):
        self.assertFalse(self.corpus.types.filter(slug='kartoffelsalad').exists())
        self.client.force_login(self.user)
        request = self.make_create_request(
            corpus=str(self.corpus.id),
            elt_type='kartoffelsalad',
            name='something',
        )
        with self.assertNumQueries(4):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            'type': ["Type 'kartoffelsalad' not found in corpus Unit Tests"],
        })

    def test_create_image_not_checked(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id)
        )
        self.image.status = S3FileStatus.Error
        self.image.save()
        response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertEqual(element.zone.image.status, S3FileStatus.Error)

    def test_create_image_zero_dimensions(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id)
        )
        self.image.width = 0
        self.image.height = 0
        self.image.status = S3FileStatus.Unchecked
        self.image.save()
        with self.assertNumQueries(6):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {'image': ['This image does not have valid dimensions.']})

    def test_create_element_slim_output_simple(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='slim output !',
        )
        request['path'] += '?slim_output=true'
        with self.assertNumQueries(5):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertDictEqual(response.json(), {
            "id": str(element.id),
        })

    def test_create_element_slim_output(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name='slim output !',
            parent=str(self.vol.id),
            elt_type='page',
            image=str(self.image.id),
        )
        request['path'] += '?slim_output=true'
        with self.assertNumQueries(14):
            response = self.client.post(**request)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()['id'])
        self.assertDictEqual(response.json(), {
            "id": str(element.id),
        })
